#ifndef KOREAN_H
#define KOREAN_H

#define NR_OF_ARGUMENTS 4

#define MIN(x, y) ((y) < (x) ? (y) : (x))
#define MAX(x, y) ((y) > (x) ? (y) : (x))

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/stat.h>

#include "csrc/config.h"
#include "csrc/image.h"

/* a struct to capture the input data which is needed all along the process */
typedef struct InputArgs InputArgs;
struct InputArgs
{
	int cellWidth, cellHeight;
	int verCells, horCells;
	int imgWidth, imgHeight, imgDim;
	int groupSize;
};

/* convert a number of given base into a binary code */
unsigned char *convertToBinary(int num, int base);

/* initialze a 'InputArgs' struct with the given input data */
void initializeInputArgs (Image *img, int cell_width, int cell_height, int group_size);

/* slit the given image to the appropriate dimensions */
Image *slitImage (Image *image);

/* calculate all the means of the grayscale values for a given cell wdth and height */
unsigned char *calculateMeans (Image *image);

/* calculated the cumulative sums for all calculated means */
short int * calculateCumSums (unsigned char *means);

/* calculate the resulting iris code for the given cumulative sums */
unsigned char *calculateIrisCode (short int *cum_sums, int post);

/* print the resulting iriscode */
void printIrisCode(unsigned char *irisCode);

/* write the result into a '.code' */
void createCodeFile(char *filename, unsigned char *irisCode);

/* the whole algorithm which just calls the above functions and returns the iriscode */
unsigned char *algoKo (char const *filename, int cell_length, int cell_height, int group_size, int post);

/* print the groups of calculated sums */
void printGroups(short int *sums);

/* print the calculated means */
void printMeans(unsigned char *means);

/* return the dimension of the result */
int getImgDim(void);

unsigned char *binaryReadFile(char *filename, int filesize);

int getFileSize(char *filename);


#endif
